from typing import TYPE_CHECKING, Any, Dict

from irckaaja.protocol import User

if TYPE_CHECKING:
    from irckaaja.client import IrcClient  # pragma: no cover


class BotScript:
    """
    Abstract script class. Should be inherited by script classes.
    See HelloWorld in scripts.helloworld.py for example.
    """

    def __init__(self, client: "IrcClient", config: Dict[str, Any]) -> None:
        self.client = client
        self.config = config
        self.alive = True

        # usage: self.say(target, message)
        self.say = client.send_privmsg

        # usage: self.joinChannel(channel_name)
        self.join_channel = client.join_channel

        # usage: self.partChannel(channel_name, reason = "")
        self.leave_channel = client.leave_channel

    def kill(self) -> None:
        self.alive = False

    # Implement methods below in the subclass to subscribe to events.
    def on_channel_message(
        self, user: User, channel_name: str, message: str
    ) -> None:
        """
        Called when a channel message is received.
        """

    def on_private_message(self, user: User, message: str) -> None:
        """
        Called when a private message is received.
        """

    def on_join(self, user: User, channel_name: str) -> None:
        """
        Called when a user joins a channel.
        """

    def on_part(self, user: User, channel_name: str) -> None:
        """
        Called when a user parts a channel.
        """

    def on_quit(self, user: User) -> None:
        """
        Called when a user quits the network.
        """

    def on_connect(self) -> None:
        """
        Called when bot has connected to the network.
        """

    def on_topic(self, user: User, channel_name: str, topic: str) -> None:
        """
        Called when topic has changed on a channel.
        """



import time
from threading import Thread
from typing import Dict, List, Optional, cast

from irckaaja.channel import IrcChannel
from irckaaja.config import BotConfig, ScriptConfig, ServerConfig
from irckaaja.connection import IrcConnection
from irckaaja.dynamicmodule import DynamicModule
from irckaaja.protocol import (
    ChannelMessage,
    CTCPVersionMessage,
    EndOfMotdMessage,
    JoinMessage,
    MessageType,
    ParsedMessage,
    PartMessage,
    PingMessage,
    PrivateMessage,
    QuitMessage,
    TopicMessage,
    TopicReplyMessage,
    User,
    UsersEndMessage,
    UsersMessage,
)


class IrcClient:
    """
    IRC Client
    """

    PING_INTERVAL_THRESHOLD = 300  # 300 seconds

    def __init__(
        self,
        server_config: ServerConfig,
        bot_config: BotConfig,
        modules_config: Dict[str, ScriptConfig],
    ) -> None:
        self.alive = True
        self.connected = False
        self.server_config = server_config
        self.bot_config = bot_config
        self.networkname = server_config.name

        self._reader_thread = Thread(target=self._connection_loop)
        self._connection = IrcConnection(
            self.server_config.hostname,
            self.server_config.port,
            self.server_config.use_tls,
            self.server_config.ca_path,
        )

        self._channel_list: List[IrcChannel] = []

        self.modules_config = modules_config
        self.dynamic_modules: List[DynamicModule] = [
            DynamicModule(self, script_config.module_name, script_config.config)
            for script_config in modules_config.values()
        ]

        self._last_ping = time.monotonic()

    def _route_message(self, parsed: ParsedMessage) -> None:
        type = parsed.type

        if parsed.message is None:
            self._unknown_message_received(parsed.raw_message or "")
            return

        if type == MessageType.PRIVATE_MESSAGE:
            self._private_message_received(cast(PrivateMessage, parsed.message))
        elif type == MessageType.JOIN:
            self._join_received(cast(JoinMessage, parsed.message))
        elif type == MessageType.PART:
            self._part_received(cast(PartMessage, parsed.message))
        elif type == MessageType.PING:
            self._ping_received(cast(PingMessage, parsed.message))
        elif type == MessageType.QUIT:
            self._quit_received(cast(QuitMessage, parsed.message))
        elif type == MessageType.TOPIC:
            self._topic_received(cast(TopicMessage, parsed.message))
        elif type == MessageType.END_OF_MOTD:
            self._motd_received(cast(EndOfMotdMessage, parsed.message))
        elif type == MessageType.TOPIC_REPLY:
            self._topic_reply_received(cast(TopicReplyMessage, parsed.message))
        elif type == MessageType.USERS:
            self._users_received(cast(UsersMessage, parsed.message))
        elif type == MessageType.END_OF_USERS:
            self._users_end_received(cast(UsersEndMessage, parsed.message))
        elif type == MessageType.CHANNEL_MESSAGE:
            self._channel_message_received(cast(ChannelMessage, parsed.message))
        elif type == MessageType.CTCP_VERSION:
            self._ctcp_version_received(
                cast(CTCPVersionMessage, parsed.message)
            )
        elif type in [
            MessageType.CTCP_TIME,
            MessageType.CTCP_VERSION,
            MessageType.CTCP_PING,
            MessageType.CTCP_DCC,
        ]:
            self._ctcp_message_received()

    def connect(self) -> None:
        """
        Non-blocking method that initiates connection.
        """
        self._reader_thread = Thread(target=self._connection_loop)
        self._reader_thread.start()

    def _connect(self) -> None:
        while self.alive:
            try:
                self._connection.close()
                self._connection = IrcConnection(
                    self.server_config.hostname,
                    self.server_config.port,
                    self.server_config.use_tls,
                    self.server_config.ca_path,
                )
                self._connection.connect()

                self.set_nick(self.bot_config.nick)
                self.set_user(
                    self.bot_config.username, self.bot_config.realname
                )

                self._last_ping = time.monotonic()
                break

            except Exception as e:
                self._print_line(str(e) + " " + self.server_config.hostname)
                self._print_line("Trying again in 30 seconds.")
                self._sleep(30)

    def _connection_loop(self) -> None:
        while self.alive:
            self._connect()
            self._process()

            if not self.alive:
                break

            self._print_line("Trying again in 60 seconds.")
            self._sleep(60)

    def _write(self, message: str) -> None:
        """
        Prints and writes message to server.
        """
        self._print_line(message[:-1])
        self._connection.write(message)

    def _check_ping_time(self) -> bool:
        return (
            time.monotonic() - self._last_ping
            < IrcClient.PING_INTERVAL_THRESHOLD
        )

    def _process(self) -> None:
        """
        Reads and handles messages.
        """
        while self.alive and self._check_ping_time():
            try:
                messages = self._connection.read()
            except (OSError, EOFError) as e:
                self._print_line(str(e))
                break

            if not self.alive:
                break

            self._handle_messages(messages)

        self._connection.close()
        self._print_line("Connection closed.")
        self.connected = False

    def _handle_messages(self, messages: List[ParsedMessage]) -> None:
        """
        Handles a list of messages
        """
        for message in messages:
            self._route_message(message)

    def _print_line(self, message: str) -> None:
        """
        Prints message with timestamp.
        """
        print(
            time.strftime("%H:%M:%S") + " |" + self.networkname + "| " + message
        )

    def set_nick(self, nick: str) -> None:
        """
        Sets user's nick on server.
        """
        self._write("NICK " + nick + "\r\n")

    def set_user(self, username: str, realname: str) -> None:
        """
        Sets username and realname to server on connect.
        """
        self._write("USER " + username + " 0 * :" + realname + "\r\n")

    def send_pong(self, message: str) -> None:
        """
        Reply to PING.
        """
        self._write("PONG :" + message + "\r\n")

    def join_channel(self, channel: str) -> None:
        """
        Joins a irc channel.
        """
        self._write("JOIN :" + channel + "\r\n")

    def leave_channel(self, channel: str, reason: str = "") -> None:
        """
        PARTs from a channel.
        """
        msg = "PART " + channel
        if reason:
            msg += " :" + reason
        self._write(msg + "\r\n")

    def send_privmsg(self, target: str, message: str) -> None:
        """
        Sends PRIVMSG to target.
        """
        self._write("PRIVMSG " + target + " :" + message + "\r\n")

    def send_ping(self, message: str) -> None:
        """
        Sends PING to server.
        """
        self._write("PING " + message + "\r\n")

    def send_CTCP(self, target: str, message: str) -> None:
        self.send_privmsg(target, str("\x01" + message + "\x01"))

    def send_notice(self, target: str, message: str) -> None:
        self._write("NOTICE " + target + " :" + message + "\r\n")

    def _on_connect(self) -> None:
        """
        Called when connected to the network.
        """
        self.send_ping(self.server_config.hostname)
        self._join_channels()

        for dm in self.dynamic_modules:
            try:
                dm.instance.on_connect()
            except Exception as e:
                print(e)

    def _join_channels(self) -> None:
        """
        Joins channels specified in server_config.
        """
        for channel in self.server_config.channels:
            self.join_channel(channel)

    def kill(self) -> None:
        """
        Called when the thread is wanted dead.
        """
        self.alive = False
        for m in self.dynamic_modules:
            m.instance.kill()

    def _private_message_received(self, msg: PrivateMessage) -> None:
        """
        Called when a private message has been received. Prints it
        and calls on_private_message() on BotScript instances.
        """
        source = msg.source.nick
        message = msg.message
        self._print_line("PRIVATE" + " <" + source + "> " + message)

        for dm in self.dynamic_modules:
            try:
                dm.instance.on_private_message(msg.source, message)
            except Exception as e:
                print(e)

    def _channel_message_received(self, msg: ChannelMessage) -> None:
        """
        Called when a PRIVMSG to a channel has been received. Prints it
        and calls on_channel_message() on BotScript instances.
        """
        source = msg.source.nick
        message = msg.message
        msg.source.full_mask
        channel = msg.channel

        self._print_line(channel + " <" + source + "> " + message)

        for dm in self.dynamic_modules:
            try:
                dm.instance.on_channel_message(msg.source, channel, message)
            except Exception as e:
                print(e)

    def _ping_received(self, msg: PingMessage) -> None:
        """
        Called when PING message has been received.
        """

        self._last_ping = time.monotonic()
        message = msg.message
        self.send_pong(message)

    def _motd_received(self, msg: EndOfMotdMessage) -> None:
        """
        Called when the end of MOTD message
        has been received.
        """
        message = msg.message

        self._print_line(message)
        if not self.connected:
            self.connected = True
            self._on_connect()

    def _find_channel_by_name(self, channel_name: str) -> Optional[IrcChannel]:
        """
        Returns a channel instance from channel_list
        matching channel_name parameter or None.
        """
        for channel in self._channel_list:
            if channel.name == channel_name:
                return channel
        return None

    def _add_channel(self, name: str, user_list: List[str]) -> None:
        """
        Adds a channel to networks channel list.
        """
        if self._find_channel_by_name(name):
            return

        channel = IrcChannel(name, user_list)
        self._channel_list.append(channel)

    def _users_received(self, msg: UsersMessage) -> None:
        """
        Called when USERS message is received. Notifies
        channel instance of the users.
        """

        channel_name = msg.channel
        user_list = msg.users

        channel = self._find_channel_by_name(channel_name)
        if not channel:
            self._add_channel(channel_name, user_list)
            return

        channel.users_message(user_list)

    def _users_end_received(self, msg: UsersEndMessage) -> None:
        """
        Called when USERS message's end has been received.
        Notifies the channel instance.
        """

        channel_name = msg.channel

        channel = self._find_channel_by_name(channel_name)
        if not channel:
            # TODO FIX
            self._print_line("REPORT THIS: usersEndReceived, channel not found")
            return

        channel.users_message_end()
        self._print_line("USERS OF " + channel_name)
        self._print_line(" ".join(channel.userlist))

    def _quit_received(self, msg: QuitMessage) -> None:
        """
        Called when a QUIT message has been received. Calls
        on_quit() on BotScripts
        """

        for channel in self._channel_list:
            channel.remove_user(msg.user.nick)

        self._print_line(msg.user.nick + " has quit.")

        for dm in self.dynamic_modules:
            try:
                dm.instance.on_quit(msg.user)
            except Exception as e:
                print(e)

    def _part_received(self, msg: PartMessage) -> None:
        """
        Called when a PART message has been received. Calls
        on_part() on BotScripts
        """

        channel_name = msg.channel
        nick = msg.user.nick
        msg.user.full_mask

        channel = self._find_channel_by_name(channel_name)
        if not channel:
            return

        channel.remove_user(nick)

        self._print_line(nick + " has parted " + channel_name)

        for dm in self.dynamic_modules:
            try:
                dm.instance.on_part(msg.user, channel_name)
            except Exception as e:
                print(e)

    def _join_received(self, msg: JoinMessage) -> None:
        """
        Called when a JOIN message has been received. Calls
        on_join() on BotScripts
        """

        nick = msg.user.nick
        channel_name = msg.channel
        msg.user.full_mask

        channel = self._find_channel_by_name(channel_name)
        if channel:
            channel.add_user(nick)

        self._print_line(nick + " has joined " + channel_name)
        for dm in self.dynamic_modules:
            try:
                dm.instance.on_join(msg.user, channel_name)
            except Exception as e:
                print(e)

    def _topic_received(self, msg: TopicMessage) -> None:
        """
        Called when topic is changed on a channel. Calls on_topic()
        on BotScripts
        """

        nick = msg.user.nick
        channel_name = msg.channel
        topic = msg.topic

        channel = self._find_channel_by_name(channel_name)
        if channel:
            channel.topic = topic

        self._print_line(
            nick + " changed the topic of " + channel_name + " to: " + topic
        )
        for dm in self.dynamic_modules:
            try:
                dm.instance.on_topic(msg.user, channel_name, topic)
            except Exception as e:
                print(e)

    def _topic_reply_received(self, msg: TopicReplyMessage) -> None:
        """
        Called when server responds to client's /topic or server informs
        of the topic on joined channel.
        """

        channel_name = msg.channel
        topic = msg.topic

        channel = self._find_channel_by_name(channel_name)
        if channel:
            channel.topic = topic

        self._print_line("Topic in " + channel_name + ": " + topic)
        for dm in self.dynamic_modules:
            try:
                dm.instance.on_topic(User(msg.nick, ""), channel_name, topic)
            except Exception as e:
                print(e)

    def _ctcp_message_received(
        self,
    ) -> None:
        self._print_line("CTCP")

    def _ctcp_version_received(self, msg: CTCPVersionMessage) -> None:
        ":juke!juke@jukk.is NOTICE irckaaja :VERSION ?l? hakkeroi!"
        self.send_notice(msg.user.nick, "\x01VERSION irckaaja 0.1.0\x01")

    def _unknown_message_received(self, msg: str) -> None:
        self._print_line(msg)

    def _sleep(self, seconds: float) -> None:
        """
        Sleeps for seconds unless not self.alive.
        """
        start = time.monotonic()
        while time.monotonic() < start + seconds and self.alive:
            time.sleep(1)



from typing import List


class IrcChannel:
    """
    Channel datastructure. Keeps track of users.
    """

    def __init__(self, name: str, userlist: List[str]) -> None:
        self.name = name
        self.userlist = userlist
        self.userlist_complete = True
        self.topic = ""

    def users_message(self, userlist: List[str]) -> None:
        """
        Usually received on join. Count of these
        messages depend on user count on the channel.
        """
        if self.userlist_complete:
            self.userlist = userlist
            self.userlist_complete = False
        else:
            self.userlist += userlist

    def users_message_end(self) -> None:
        """
        When received, states that all users on channel
        have been listed and userlist is therefore complete.
        """
        self.userlist_complete = True

    def remove_user(self, nick: str) -> None:
        """
        Tries to remove a user from the channel.
        """
        try:
            self.userlist.remove(nick)
        except ValueError:
            pass

    def add_user(self, nick: str) -> None:
        """
        Adds user to channel if not already added.
        """
        if nick not in self.userlist:
            self.userlist.append(nick)



from dataclasses import dataclass, field
from enum import IntEnum
from typing import List, Optional, Tuple


class MessageType(IntEnum):
    PRIVATE_MESSAGE = 0
    JOIN = 1
    PART = 2
    PING = 3
    QUIT = 4
    TOPIC = 5
    END_OF_MOTD = 6
    NICK_IN_USE = 7
    TOPIC_REPLY = 9
    USERS = 10
    END_OF_USERS = 11
    CHANNEL_MESSAGE = 12
    CTCP_VERSION = 13
    CTCP_PING = 14
    CTCP_TIME = 15
    CTCP_DCC = 16
    UNKNOWN = 255


@dataclass
class User:
    nick: str
    full_mask: str


class Message:
    pass


@dataclass
class PrivateMessage(Message):
    source: User
    message: str


@dataclass
class ChannelMessage(Message):
    source: User
    message: str
    channel: str


@dataclass
class JoinMessage(Message):
    user: User
    channel: str


@dataclass
class PartMessage(Message):
    user: User
    channel: str


@dataclass
class QuitMessage(Message):
    user: User
    message: Optional[str]


@dataclass
class TopicMessage(Message):
    user: User
    channel: str
    topic: str


@dataclass
class TopicReplyMessage(Message):
    nick: str
    channel: str
    topic: str


@dataclass
class PingMessage(Message):
    message: str


@dataclass
class PongMessage(Message):
    message: str


@dataclass
class UsersMessage(Message):
    channel: str
    users: List[str]


@dataclass
class UsersEndMessage(Message):
    channel: str


@dataclass
class EndOfMotdMessage(Message):
    message: str


@dataclass
class CTCPVersionMessage(Message):
    user: User


@dataclass
class CTCPPingMessage(Message):
    user: User
    id: str
    time: str


@dataclass
class CTCPTimeMessage(Message):
    user: User


@dataclass
class CTCPDCCMessage:
    user: User


@dataclass
class ParsedMessage:
    type: MessageType
    message: Optional[Message] = None

    raw_message: Optional[str] = None
    prefix: Optional[str] = None
    command: Optional[str] = None
    params: List[str] = field(default_factory=list)


def stripping_partition(s: str) -> Tuple[str, str]:
    """
    <SPACE>    ::= ' ' { ' ' }
    """
    before, _, rest = s.partition(" ")
    return before, rest.lstrip(" ")


def parse_full_mask(mask: str) -> User:
    """
    <prefix>   ::= <servername> | <nick> [ '!' <user> ] [ '@' <host> ]
    """

    if "!" in mask:
        nick, _, _ = mask.partition("!")
        return User(nick=nick, full_mask=mask)
    if "@" in mask:
        nick, _, _ = mask.partition("@")
        return User(nick=nick, full_mask=mask)

    return User(nick=mask, full_mask=mask)


@dataclass
class Atoms:
    prefix: str
    command: str
    params: List[str]

    @classmethod
    def parse_params(cls, params_raw: str) -> List[str]:
        """
        <params>   ::= <SPACE> [ ':' <trailing> | <middle> <params> ]
        <middle>   ::= <Any *non-empty* sequence of octets not including SPACE
                    or NUL or CR or LF, the first of which may not be ':'>
        <trailing> ::= <Any, possibly *empty*, sequence of octets not including
                        NUL or CR or LF>
        """
        params = []
        rest = params_raw
        while rest != "":
            if rest.startswith(":"):
                # trailing i.e. last param
                params.append(rest[1:])
                break
            param, rest = stripping_partition(rest)
            params.append(param)
        return params

    @classmethod
    def from_message(cls, message: str) -> "Atoms":
        """
        From https://datatracker.ietf.org/doc/html/rfc1459#section-2.3.1

        <message>  ::= [':' <prefix> <SPACE> ] <command> <params> <crlf>
        <prefix>   ::= <servername> | <nick> [ '!' <user> ] [ '@' <host> ]
        <command>  ::= <letter> { <letter> } | <number> <number> <number>
        <SPACE>    ::= ' ' { ' ' }
        <params>   ::= <SPACE> [ ':' <trailing> | <middle> <params> ]

        <middle>   ::= <Any *non-empty* sequence of octets not including SPACE
                    or NUL or CR or LF, the first of which may not be ':'>
        <trailing> ::= <Any, possibly *empty*, sequence of octets not including
                        NUL or CR or LF>

        <crlf>     ::= CR LF
        """
        rest = message

        prefix = ""
        if rest.startswith(":"):
            prefix, rest = stripping_partition(rest)
            prefix = prefix[1:]

        command, params_raw = stripping_partition(rest)
        params = cls.parse_params(params_raw)

        return Atoms(prefix=prefix, command=command, params=params)


class MessageParser:
    def __init__(self) -> None:
        self.parsers = {
            "PRIVMSG": self.parse_privmsg,
            "JOIN": self.parse_join,
            "PART": self.parse_part,
            "PING": self.parse_ping,
            "QUIT": self.parse_quit,
            "353": self.parse_users,
            "366": self.parse_users_end,
            "376": self.parse_end_of_motd,
            "TOPIC": self.parse_topic,
            "332": self.parse_server_topic,
        }

    def parse_buffer(self, buff: str) -> Tuple[List[ParsedMessage], str]:
        """
        Parses buffer to ParsedMessages.
        Returns list of ParsedMessages and remainder of buff.
        """
        parsed_messages = []
        while "\r\n" in buff:
            message, _, buff = buff.partition("\r\n")
            parsed = self.parse_message(message)
            if parsed is None:
                continue
            parsed_messages.append(parsed)

        return parsed_messages, buff

    def parse_message(self, message: str) -> Optional[ParsedMessage]:
        atoms = Atoms.from_message(message)
        if atoms.command in self.parsers:
            return self.parsers[atoms.command](atoms.prefix, atoms.params)

        return ParsedMessage(
            MessageType.UNKNOWN,
            raw_message=message,
            prefix=atoms.prefix,
            command=atoms.command,
            params=atoms.params,
        )

    def parse_privmsg(
        self, source_full_mask: str, params: List[str]
    ) -> ParsedMessage:
        source = parse_full_mask(source_full_mask)
        target, message = params
        if target.startswith("#") or target.startswith("!"):
            return ParsedMessage(
                type=MessageType.CHANNEL_MESSAGE,
                message=ChannelMessage(
                    source=source,
                    channel=target,
                    message=message,
                ),
            )
        return ParsedMessage(
            type=MessageType.PRIVATE_MESSAGE,
            message=PrivateMessage(
                source=source,
                message=message,
            ),
        )

    def parse_join(self, full_mask: str, params: List[str]) -> ParsedMessage:
        channel = params[0]
        return ParsedMessage(
            type=MessageType.JOIN,
            message=JoinMessage(
                user=parse_full_mask(full_mask), channel=channel
            ),
        )

    def parse_part(self, full_mask: str, params: List[str]) -> ParsedMessage:
        channel = params[0]
        return ParsedMessage(
            type=MessageType.PART,
            message=PartMessage(
                user=parse_full_mask(full_mask), channel=channel
            ),
        )

    def parse_ping(self, _: str, params: List[str]) -> ParsedMessage:
        message = params[0]
        return ParsedMessage(
            type=MessageType.PING, message=PingMessage(message=message)
        )

    def parse_quit(self, full_mask: str, params: List[str]) -> ParsedMessage:
        user = parse_full_mask(full_mask)
        message = params[0]

        return ParsedMessage(
            type=MessageType.QUIT,
            message=QuitMessage(user=user, message=message),
        )

    def parse_users(self, _: str, params: List[str]) -> ParsedMessage:
        """
        :example.org 353 irckaaja @ #channelname :yournick @juke
        """
        _, _, channel, raw_user_list = params
        user_list = raw_user_list.split(" ")
        return ParsedMessage(
            type=MessageType.USERS,
            message=UsersMessage(channel=channel, users=user_list),
        )

    def parse_users_end(self, _: str, params: List[str]) -> ParsedMessage:
        """
        :example.org 366 irckaaja #channelname :End of /NAMES list.
        :another.example.org 366 irckaaja @ #channelname :End of /NAMES list.
        """
        if len(params) == 3:
            _, channel, _ = params
        else:
            _, _, channel, _ = params

        return ParsedMessage(
            type=MessageType.END_OF_USERS,
            message=UsersEndMessage(channel=channel),
        )

    def parse_end_of_motd(self, _: str, params: List[str]) -> ParsedMessage:
        """
        :example.org 376 irckaaja :End of /MOTD command.
        """
        message = params[0]
        return ParsedMessage(
            type=MessageType.END_OF_MOTD,
            message=EndOfMotdMessage(message=message),
        )

    def parse_topic(self, full_mask: str, params: List[str]) -> ParsedMessage:
        """
        :user!~user@example.org TOPIC #channelname :topic message
        """
        user = parse_full_mask(full_mask)
        channel, topic = params
        return ParsedMessage(
            type=MessageType.TOPIC,
            message=TopicMessage(
                user=user,
                channel=channel,
                topic=topic,
            ),
        )

    def parse_server_topic(self, _: str, params: List[str]) -> ParsedMessage:
        """
        :server.example.org 332 irckaaja #testchannel :let's test
        """
        nick, channel, topic = params
        return ParsedMessage(
            type=MessageType.TOPIC_REPLY,
            message=TopicReplyMessage(
                nick=nick,
                channel=channel,
                topic=topic,
            ),
        )



import importlib
from typing import TYPE_CHECKING, Any, Dict, Optional

from irckaaja.botscript import BotScript

if TYPE_CHECKING:
    from irckaaja.client import IrcClient  # pragma: no cover


class DynamicModule:
    """
    Container for reloadable scripts.
    """

    def __init__(
        self,
        connection: "IrcClient",
        module_name: str,
        config: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Initialises and tries to load a class in a module in the scripts folder.
        Module should be named <ClassName>.lower().

        connection: connection to the network in which the module is
        related
        modulename: name of the module
        classvar: script class
        instance: instance of classvar
        """
        self.connection = connection
        self.module_name = module_name
        self.module_config = config

        self.module = __import__(
            "irckaaja.scripts." + self.module_name.lower(),
            None,
            None,
            [self.module_name],
            0,
        )

        self.classvar = getattr(self.module, self.module_name)
        self.instance: BotScript = self.classvar(
            self.connection, self.module_config
        )

    def reload_module(self) -> None:
        """
        Reloads the module, the class and overwrites the instance.
        """
        self.instance.kill()
        importlib.reload(self.module)
        self.classvar = getattr(self.module, self.module_name)
        self.instance = self.classvar(self.connection, self.module_config)



from irckaaja.botscript import BotScript, User


class HelloWorld(BotScript):
    """
    A simple script class. Only responds to messages starting "hello"
    in every joined channel.
    """

    def on_channel_message(
        self, user: User, channel_name: str, message: str
    ) -> None:
        if message.startswith("hello"):
            self.say(channel_name, "hello, " + user.nick)
            return



import socket
import ssl
from datetime import timedelta
from typing import List, Optional

from irckaaja.protocol import MessageParser, ParsedMessage


class IrcConnection:
    """
    IRC Connection
    """

    def __init__(
        self,
        hostname: str,
        port: int,
        use_tls: bool,
        cafile: Optional[str] = None,
        timeout: timedelta = timedelta(seconds=1),
    ) -> None:
        self._hostname = hostname
        self._port = port
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._use_tls = use_tls
        self._cafile = cafile
        self._buff = ""
        self._parser = MessageParser()
        self._encoding = "utf-8"
        self._timeout = timeout.total_seconds()

    def connect(self) -> None:
        self._socket.connect((self._hostname, self._port))
        if self._use_tls:
            ssl_context = ssl.create_default_context(cafile=self._cafile)
            self._socket = ssl_context.wrap_socket(
                self._socket, server_hostname=self._hostname
            )
        self._socket.settimeout(self._timeout)

    def write(self, message: str) -> None:
        self._socket.sendall(bytearray(message, self._encoding))

    def read(self) -> List[ParsedMessage]:
        try:
            read = self._socket.recv(4096)
        except socket.timeout:
            return []

        if len(read) == 0:
            # socket.recv() returns empty buff for closed sockets
            raise EOFError()

        self._buff += read.decode(self._encoding)
        parsed_messages, self._buff = self._parser.parse_buffer(self._buff)
        return parsed_messages

    def close(self) -> None:
        self._socket.close()



from optparse import OptionParser, Values
from time import sleep

from irckaaja.client import IrcClient
from irckaaja.config import Config


def get_options() -> Values:
    parser = OptionParser()
    parser.add_option(
        "-c",
        "--configfile",
        dest="configfile",
        help="Reads FILE as a configfile",
        metavar="FILE",
        default="config.ini",
    )

    options, _ = parser.parse_args()
    return options


def main() -> None:
    options = get_options()
    conf = Config(options.configfile)
    modulesd = conf.modules()
    bot_info = conf.bot()
    connection_list = []

    for server_conf in conf.servers().values():
        connection_list.append(IrcClient(server_conf, bot_info, modulesd))

    for s in connection_list:
        s.connect()

    # Interrupts are only handled in the main thread in Python so...
    while True:
        try:
            sleep(1.0)
        except KeyboardInterrupt:
            for s in connection_list:
                s.kill()
            break


if __name__ == "__main__":
    main()



__author__ = "juke"
import re
import urllib
from typing import Any, List
from xml.dom import minidom

from irckaaja.botscript import BotScript, User
from irckaaja.scripts.urls import parse_urls

APIURL = "http://gdata.youtube.com/feeds/api/videos/"


class YoutubeAnnouncer(BotScript):
    def on_channel_message(
        self, user: User, channel_name: str, message: str
    ) -> None:
        ids = self._parse_ids(message)

        if ids:
            for id in ids:
                title = self._query_title(id)
                if title:
                    self.say(channel_name, "** " + title + " **")

    @staticmethod
    def _parse_ids(message: str) -> List[str]:
        urls = parse_urls(message)
        if not urls:
            return []
        ids = []
        for url in urls:
            match = re.match(
                re.compile(
                    r"""https?://
                        (youtu\.be/|www\.youtube\.com/watch\?v=)
                        (\S*?)($|\?\S*$)
                         """,
                    re.X,
                ),
                url,
            )

            if match:
                ids.append(match.group(2))
        return ids

    @staticmethod
    def _query_title(id: str) -> Any:
        xml = urllib.request.urlopen(APIURL + id).read()
        return (
            minidom.parseString(xml)
            .getElementsByTagName("title")[0]
            .childNodes[0]
            .nodeValue
        )



import dbm
import time
from typing import Any, Dict, Tuple

from irckaaja.botscript import BotScript, User
from irckaaja.client import IrcClient
from irckaaja.scripts.urls import parse_urls


def serialize(ts: float, nick: str, message: str) -> bytes:
    return ",".join([str(int(ts)), nick, message]).encode()


def deserialize(b: bytes) -> Tuple[float, str, str]:
    ts, nick, message = b.decode().split(",", 2)
    return float(ts), nick, message


class IrcLinkHistory(BotScript):
    def __init__(self, client: IrcClient, config: Dict[str, Any]) -> None:
        BotScript.__init__(self, client, config)

        self.store_path = config["store_path"] or "./"

        self.channels = (
            [config["channels"]]
            if isinstance(config["channels"], str)
            else config["channels"]
        )
        self.dbs: Dict[str, dbm._Database] = {}

        for channel in self.channels:
            assert channel, "Channel name must be non empty"
            self.dbs[channel] = dbm.open(self.store_path + "/" + channel, "c")

    def _get_diff_string(self, t1: float, t2: float) -> str:
        diff = t1 - t2

        if diff < 60:
            return "%d seconds" % diff
        elif diff < 60 * 60:
            minutes = diff / 60.0
            seconds = (minutes - int(minutes)) * 60
            return "%d minutes, %d seconds" % (minutes, seconds)
        elif diff < 60 * 60 * 24:
            hours = diff / (60.0 * 60)
            minutes = (hours - int(hours)) * 60
            seconds = (minutes - int(minutes)) * 60
            return "%d hours, %d minutes, %d seconds" % (
                hours,
                minutes,
                seconds,
            )
        elif diff < 60 * 60 * 24 * 365:
            days = diff / (60.0 * 60 * 24)
            hours = (days - int(days)) * 24
            minutes = (hours - int(hours)) * 60
            seconds = (minutes - int(minutes)) * 60
            return "%d days, %d hours, %d minutes, %d seconds" % (
                days,
                hours,
                minutes,
                seconds,
            )
        else:
            years = diff / (60.0 * 60 * 24 * 365)
            days = (years - int(years)) * 365
            hours = (days - int(days)) * 24
            minutes = (hours - int(hours)) * 60
            seconds = (minutes - int(minutes)) * 60
            return "%d years, %d days, %d hours, %d minutes, %d seconds" % (
                years,
                days,
                hours,
                minutes,
                seconds,
            )

    def on_channel_message(
        self,
        user: User,
        channel_name: str,
        message: str,
    ) -> None:
        urls = parse_urls(message)

        if not urls:
            return
        try:
            db = self.dbs[channel_name]
        except KeyError:
            return

        for url in urls:
            url = url.strip()
            history_tuple_raw = db.get(url)
            if not history_tuple_raw:
                db[url] = serialize(time.time(), user.nick, message)
            else:
                old_time, old_nick, old_message = deserialize(history_tuple_raw)
                self.say(
                    channel_name,
                    "previously seen "
                    + self._get_diff_string(time.time(), old_time)
                    + " ago",
                )
                self.say(
                    channel_name, "< " + old_nick + "> " + old_message + ""
                )



from dataclasses import dataclass
from typing import Any, Dict, Optional

from configobj import ConfigObj

CONFIG_FILENAME = "config.ini"


@dataclass
class ServerConfig:
    name: str
    channels: list[str]
    hostname: str
    port: int
    use_tls: bool
    ca_path: Optional[str] = None


@dataclass
class BotConfig:
    nick: str
    altnick: str
    realname: str
    username: str
    owner: str


@dataclass
class ScriptConfig:
    module_name: str
    config: Dict[str, Any]


class Config:
    """
    Wrapper for config.ini
    """

    def __init__(self, config_filename: str = CONFIG_FILENAME) -> None:
        self.filename = config_filename
        self.config = ConfigObj(
            self.filename, list_values=True, encoding="utf-8"
        )

    def servers(self) -> Dict[str, ServerConfig]:
        """
        Returns servers as a dictionary.
        """
        config = {}
        for alias, server in self.config["servers"].items():
            channels_raw = server.get("channels", [])
            if isinstance(channels_raw, str):
                channels = [channels_raw]
            else:
                channels = channels_raw

            use_tls_raw = server.get("use_tls", "False")
            use_tls = False
            if use_tls_raw.lower() == "true":
                use_tls = True

            config[alias] = ServerConfig(
                name=alias,
                hostname=server["hostname"],
                port=int(server.get("port", "6667")),
                channels=channels,
                use_tls=use_tls,
                ca_path=server.get("ca_path"),
            )
        return config

    def modules(self) -> Dict[str, ScriptConfig]:
        """
        Returns a dictionary of modules defined in the
        conf to be loaded.
        """
        modules = {}
        for name, config in self.config["modules"].items():
            modules[name] = ScriptConfig(module_name=name, config=config)
        return modules

    def bot(self) -> BotConfig:
        """
        Returns bot dictionary.
        """
        config = self.config["bot"]
        return BotConfig(
            nick=config["nick"],
            altnick=config.get("altnick", config["nick"] + "_"),
            realname=config["realname"],
            username=config["username"],
            owner=config["owner"],
        )



from irckaaja.botscript import BotScript
from irckaaja.dynamicmodule import DynamicModule
from irckaaja.protocol import User


class ModuleLoader(BotScript):
    """
    A helper script for loading, reloading and unloading
    modules dynamically on runtime. Makes developing scripts
    much easier as the bot doesn'tt need to reconnect to irc network
    between changes.
    """

    def on_private_message(self, source: User, message: str) -> None:
        # authenticating
        if source.full_mask != self.client.bot_config.owner:
            return

        # checking if we got a relevant message
        try:
            if self._try_load(source, message):
                return
            if self._try_reload(source, message):
                return
            if self._try_unload(source, message):
                return

        except Exception as e:
            self.say(source.nick, "error: " + str(e))

    def _try_load(self, source: User, message: str) -> bool:
        if not message.startswith("!load"):
            return False

        module_name = message.replace("!load ", "")

        # Not loading if it's already loaded
        for dm in self.client.dynamic_modules:
            if dm.module_name == module_name:
                self.say(
                    source.nick,
                    "module " + str(dm.classvar) + " already loaded",
                )
                return True

        # Loading and appending to the list
        dm = DynamicModule(
            self.client,
            module_name,
            self.client.modules_config[module_name].config,
        )
        self.client.dynamic_modules.append(dm)

        self.say(source.nick, "loaded " + str(dm.classvar))
        return True

    def _try_reload(self, source: User, message: str) -> bool:
        if not message.startswith("!reload"):
            return False

        module_name = message.replace("!reload ", "")

        # finding the module to reload
        for dm in self.client.dynamic_modules:
            if dm.module_name != module_name:
                continue

            dm.reload_module()
            self.say(source.nick, "reloaded " + str(dm.classvar))
            return True

        # didn't find it
        self.say(source.nick, "unable to find module with name " + module_name)
        return True

    def _try_unload(self, source: User, message: str) -> bool:
        if not message.startswith("!unload"):
            return False

        module_name = message.replace("!unload ", "")

        # finding the module in the list by name
        for dm in self.client.dynamic_modules:
            if dm.module_name != module_name:
                continue

            # unloading
            self.client.dynamic_modules.remove(dm)
            self.say(source.nick, "unloaded  " + str(dm.classvar))
            del dm
            return True

        # module was not found
        self.say(source.nick, "unable to find module with name " + module_name)
        return True
